@@ -1,14 +1,18 @@
+import copy
 import time
 from collections import namedtuple
 
 import traceback
+from pprint import pformat
 from typing import Union
 
+from celery import current_app
 from celery.result import AsyncResult
-from celery.signals import task_prerun
-from celery.states import FAILURE, REVOKED
+from celery.signals import before_task_publish
+from celery.states import FAILURE, REVOKED, PENDING, STARTED, RECEIVED, RETRY
 from celery.utils.log import get_task_logger
 from firexkit.broker import handle_broker_timeout
+from firexkit.inspect import get_task, get_active_queues
 from firexkit.revoke import RevokedRequests
 
 RETURN_KEYS_KEY = '__task_return_keys'
@@ -16,18 +20,37 @@ RETURN_KEYS_KEY = '__task_return_keys'
 logger = get_task_logger(__name__)
 
 
-def get_task_name_from_result(result):
+def get_task_info_from_result(result, key: str = None):
     try:
         backend = result.app.backend
     except AttributeError:
-        from celery import current_app
         backend = current_app.backend
-    name = handle_broker_timeout(backend.get, args=(str(result),), timeout=30)
-    if name is None:
-        name = ''
+
+    if key is not None:
+        info = handle_broker_timeout(backend.client.hget, args=(str(result), key), timeout=30)
     else:
-        name = name.decode()
-    return name
+        info = handle_broker_timeout(backend.client.get, args=(str(result), ), timeout=30)
+
+    if info is None:
+        info = ''
+    else:
+        info = info.decode()
+    return info
+
+
+def get_task_name_from_result(result):
+    try:
+        return get_task_info_from_result(result=result, key='name')
+    except AttributeError:
+        return get_task_info_from_result(result=result)
+
+
+def get_task_queue_from_result(result):
+    try:
+        return get_task_info_from_result(result=result, key='queue')
+    except AttributeError:
+        logger.exception('Task queue info not supported for this broker')
+        return ''
 
 
 def get_tasks_names_from_results(results):
@@ -40,11 +63,23 @@ def get_result_logging_name(result: AsyncResult, name=None):
     return '%s[%s]' % (name, result)
 
 
-# noinspection PyUnusedLocal
-@task_prerun.connect
-def populate_task_name(task_id, task, args, kwargs, **donotcare):
-    from celery import current_app
-    current_app.backend.set(task_id, task.name)
+@before_task_publish.connect()
+def populate_task_info(sender, declare, headers, **_kwargs):
+    task_info = {'name': sender}
+    try:
+        task_info.update({'queue': declare[0].name})
+    except (IndexError, AttributeError):
+        pass
+
+    current_app.backend.client.hmset(headers['id'], task_info)
+
+
+def mark_queues_ready(*queue_names: str):
+    current_app.backend.client.sadd('QUEUES', *queue_names)
+
+
+def was_queue_ready(queue_name: str):
+    return current_app.backend.client.sismember('QUEUES', queue_name)
 
 
 def is_result_ready(result: AsyncResult, timeout=None, retry_delay=0.1):
@@ -121,6 +156,65 @@ def _check_for_traceback_in_parents(result, timeout=None, retry_delay=0.1):
             return _check_for_traceback_in_parents(parent, timeout=timeout, retry_delay=retry_delay)
 
 
+def _is_worker_alive(result: AsyncResult, timeout=None, retry_delay=0.1, retries=1):
+    task_name = get_result_logging_name(result)
+    tries = 0
+
+    # NOTE: Retries for possible false negative in the case where task changes host in the small timing window
+    # between getting task state / info and checking for aliveness. Retries for broker issues are handled downstream
+    while tries <= retries:
+        state = handle_broker_timeout(lambda r: r.state, args=(result,), timeout=timeout, retry_delay=retry_delay)
+        if not state:
+            logger.debug(f'Cannot get state for {task_name}; assuming task is alive')
+            return True
+
+        if state == STARTED or state == RECEIVED:
+            # Query the worker to see if it knows about this task
+            info = handle_broker_timeout(lambda r: r.info, args=(result,), timeout=timeout, retry_delay=retry_delay)
+            hostname = info.get('hostname') if info else None
+
+            if not hostname:
+                logger.debug(f'Cannot get run info for {task_name}; assuming task is alive.'
+                             f' Info: {info}, Hostname: {hostname}')
+                return True
+
+            task_info = get_task(method_args=(result.id,), destination=(hostname,), timeout=10)
+            if task_info and any(task_info.values()):
+                return True
+
+            logger.debug(f'Task inspection for {task_name} on {hostname} returned:\n{pformat(task_info)}')
+
+        elif state == PENDING or state == RETRY:
+            # Check if task queue is alive
+            task_queue = get_task_queue_from_result(result)
+            if not task_queue:
+                logger.debug(f'Cannot get task queue for {task_name}; assuming task is alive.')
+                return True
+
+            queue_seen = was_queue_ready(queue_name=task_queue)
+            if not queue_seen:
+                logger.debug(f'Queue "{task_queue}" for {task_name} not seen yet; assuming task is alive.')
+                return True
+
+            queues = get_active_queues()
+            active_queues = {queue['name'] for node in queues.values() for queue in node} if queues else set()
+            if task_queue in active_queues:
+                return True
+
+            logger.debug(f'Active queues inspection for {task_name} on queue {task_queue} returned:\n'
+                         f'{pformat(queues)}\n'
+                         f'Active queues: {pformat(active_queues)}')
+
+        else:
+            logger.debug(f'Unknown state ({state} for task {task_name}; assuming task is alive.')
+            return True
+
+        tries += 1
+        logger.info(f'Task {task_name} is not responding to queries. Tries: {tries}')
+
+    return False
+
+
 WaitLoopCallBack = namedtuple('WaitLoopCallBack', ['func', 'frequency', 'kwargs'])
 
 
@@ -141,7 +235,11 @@ def send_block_task_states_to_caller_task(func):
 def wait_on_async_results(results,
                           max_wait=None,
                           callbacks: [WaitLoopCallBack]=tuple(),
-                          sleep_between_iterations=0.05):
+                          sleep_between_iterations=0.05,
+                          check_task_worker_frequency=900,
+                          fail_on_worker_failures=3,
+                          log_msg=True
+                          ):
     if not results:
         return
 
@@ -154,12 +252,13 @@ def wait_on_async_results(results,
     failures = []
     for result in results:
         logging_name = get_result_logging_name(result)
-        logger.debug('-> Waiting for %s to become ready' % logging_name)
+        if log_msg:
+            logger.debug('-> Waiting for %s to complete' % logging_name)
         try:
+            worker_failures = 0
             while not is_result_ready(result):
                 if RevokedRequests.instance().is_revoked(result):
-                    raise ChainRevokedException(task_id=str(result),
-                                                task_name=get_task_name_from_result(result))
+                    break
 
                 _check_for_traceback_in_parents(result, timeout=30)
 
@@ -176,9 +275,27 @@ def wait_on_async_results(results,
                     if trials % (callback.frequency / sleep_between_iterations) == 0:
                         callback.func(**callback.kwargs)
 
+                # Check for dead workers
+                if check_task_worker_frequency and fail_on_worker_failures and \
+                        round(trials % (check_task_worker_frequency / sleep_between_iterations)) == 0:
+                    alive = _is_worker_alive(result=result)
+                    if not alive:
+                        worker_failures += 1
+                        logger.warning(f'Task {get_task_name_from_result(result)} appears to be a zombie.'
+                                       f' Failures: {worker_failures}')
+                        if worker_failures >= fail_on_worker_failures:
+                            raise ChainInterruptedByZombieTaskException(task_id=str(result),
+                                                                        task_name=get_task_name_from_result(result))
+                    else:
+                        worker_failures = 0
+
             if result.state == REVOKED:
                 raise ChainRevokedException(task_id=str(result),
                                             task_name=get_task_name_from_result(result))
+            if result.state == PENDING:
+                # Pending tasks can be in revoke list. State will still be PENDING.
+                raise ChainRevokedPreRunException(task_id=str(result),
+                                                  task_name=get_task_name_from_result(result))
             if result.state == FAILURE:
                 cause = result.result if isinstance(result.result, Exception) else None
                 raise ChainInterruptedException(task_id=str(result),
@@ -191,7 +308,8 @@ def wait_on_async_results(results,
     if len(failures) == 1:
         raise failures[0]
     elif failures:
-        multi_exception = MultipleFailuresException()
+        failed_task_ids = [e.task_id for e in failures if hasattr(e, 'task_id')]
+        multi_exception = MultipleFailuresException(failed_task_ids)
         multi_exception.failures = failures
         raise multi_exception
 
@@ -205,20 +323,25 @@ def wait_on_async_results_and_maybe_raise(results, raise_exception_on_failure=Tr
 
 
 # This is a generator that returns one AsyncResult as it completes
-def wait_for_any_results(results, max_wait=None, poll_max_wait=0.1, **kwargs):
+def wait_for_any_results(results, max_wait=None, poll_max_wait=0.1, log_msg=False, **kwargs):
     if isinstance(results, AsyncResult):
         results = [results]
     start_time = time.time()
+
+    logging_names = [f'-> {get_result_logging_name(result)}' for result in results]
+
+    logger.debug('Waiting for any of the following tasks to complete:\n' + '\n'.join(logging_names))
     while len(results):
         if max_wait and max_wait < time.time() - start_time:
             raise WaitOnChainTimeoutError('Results %r were still not ready after %d seconds' % (results, max_wait))
         for result in results:
             try:
-                wait_on_async_results_and_maybe_raise([result], max_wait=poll_max_wait, **kwargs)
+                wait_on_async_results_and_maybe_raise([result], max_wait=poll_max_wait, log_msg=log_msg, **kwargs)
             except WaitOnChainTimeoutError:
                 pass
             else:
                 yield result
+                logger.debug(f'--> {get_result_logging_name(result)} completed')
                 results.remove(result)
                 break
 
@@ -248,6 +371,10 @@ class ChainRevokedException(ChainException):
         return message
 
 
+class ChainRevokedPreRunException(ChainRevokedException):
+    pass
+
+
 class ChainInterruptedException(ChainException):
     MESSAGE = "The chain has been interrupted by a failure in microservice "
 
@@ -266,14 +393,20 @@ class ChainInterruptedException(ChainException):
         return message
 
 
-class MultipleFailuresException(ChainInterruptedException):
-    MESSAGE = "The chain has been interrupted by multiple failing microservices"
+class ChainInterruptedByZombieTaskException(ChainInterruptedException):
+    def __str__(self):
+        return super().__str__() + ': (zombie task)'
 
-    def __init__(self):
+
+class MultipleFailuresException(ChainInterruptedException):
+    MESSAGE = "The chain has been interrupted by multiple failing microservices: %s"
+
+    def __init__(self, task_ids=('UNKNOWN',)):
+        self.task_ids = task_ids
         super(ChainInterruptedException, self).__init__()
 
     def __str__(self):
-        return self.MESSAGE
+        return self.MESSAGE % self.task_ids
 
 
 def get_task_results(results: dict) -> dict:
@@ -291,7 +424,7 @@ def _get_results(result: AsyncResult, return_keys_only=True, merge_children_resu
         return results
     try:
         if result.successful():
-            _results = result.result
+            _results = copy.deepcopy(result.result) if isinstance(result.result, dict) else result.result
             if _results:
                 if return_keys_only:
                     results = get_task_results(_results)
@@ -359,7 +492,7 @@ def results2tuple(results: dict, return_keys: Union[str, tuple]) -> tuple:
     return tuple(results_to_return)
 
 
-def get_results_upto_parent(result: AsyncResult, parent_id: str = None, return_keys=(), **kwargs):
+def get_results_upto_parent(result: AsyncResult, return_keys=(), parent_id: str = None, **kwargs):
     extracted_dict = {}
     node = result
     while node and node.id != parent_id:
@@ -369,7 +502,12 @@ def get_results_upto_parent(result: AsyncResult, parent_id: str = None, return_k
                 # Since we're walking up the chain, children gets precedence in case we get the same key
                 extracted_dict[k] = v
         node = node.parent
-    return results2tuple(extracted_dict, return_keys) if return_keys else extracted_dict
+
+    from firexkit.task import FireXTask
+    if not return_keys or return_keys == FireXTask.DYNAMIC_RETURN or return_keys == (FireXTask.DYNAMIC_RETURN,):
+        return extracted_dict
+    else:
+        return results2tuple(extracted_dict, return_keys)
 
 
 def disable_async_result(result: AsyncResult):
@@ -382,3 +520,14 @@ def disable_async_result(result: AsyncResult):
 
     for child in children:
         disable_async_result(child)
+
+#
+# Returns the first exception that is not a "ChainInterruptedException"
+# in the exceptions stack.
+#
+def first_non_chain_interrupted_exception(ex):
+    e = ex
+    while e.__cause__ is not None and isinstance(e, ChainInterruptedException):
+        e = e.__cause__
+    return e
+
