@@ -18,6 +18,7 @@ from types import MethodType
 from abc import abstractmethod
 from celery.app.task import Task
 from celery.local import PromiseProxy
+from celery.signals import task_prerun, task_postrun, task_revoked
 from celery.utils.log import get_task_logger, get_logger
 from firexkit.revoke import revoke_recursively
 from firexkit.bag_of_goodies import BagOfGoodies
@@ -26,6 +27,7 @@ from firexkit.result import get_tasks_names_from_results, wait_for_any_results,
     RETURN_KEYS_KEY, wait_on_async_results_and_maybe_raise, get_result_logging_name, ChainInterruptedException, \
     ChainRevokedException
 from firexkit.resources import get_firex_css_filepath, get_firex_logo_filepath, JINJA_ENV
+import time
 
 logger = get_task_logger(__name__)
 
@@ -545,6 +547,28 @@ class FireXTask(Task):
                     self._update_child_state(child_result, self._UNBLOCKED)
         return child_result
 
+    def enqueue_in_parallel(self, chains, max_parallel_chains=15, wait_for_completion=True,
+                            raise_exception_on_failure=False):
+        """ This method executes the provided list of Signatures/Chains in parallel
+        and returns the associated list of "async_result" objects.
+        The results are returned in the same order as the input Signatures/Chains."""
+        promises = []
+        scheduled = []
+        for c in chains:
+            if len(scheduled) >= max_parallel_chains:
+                # Reach the max allowed parallel chains, wait for one to complete before scheduling the next one.
+                async_res = next(wait_for_any_results(scheduled, raise_exception_on_failure=raise_exception_on_failure))
+                scheduled.remove(async_res)
+            # Schedule the next child
+            logger.debug(f'Enqueueing: {c.get_label()}')
+            promise = self.enqueue_child(c)
+            scheduled.append(promise)
+            promises.append(promise)
+        if wait_for_completion or raise_exception_on_failure:
+            # Wait for all children to complete
+            self.wait_for_specific_children(promises, raise_exception_on_failure=raise_exception_on_failure)
+        return promises
+
     def revoke_pending_children(self, **kwargs):
         pending_children = self.pending_enqueued_children
         if pending_children:
@@ -612,13 +636,14 @@ class FireXTask(Task):
         return worker_log_url
 
     def write_task_log_html_header(self):
+        base_dir = self.task_logging_dirpath
         html_header = JINJA_ENV.get_template('log_template.html').render(
-            firex_stylesheet=get_firex_css_filepath(self.app.conf.resources_dir),
-            logo=get_firex_logo_filepath(self.app.conf.resources_dir),
+            firex_stylesheet=get_firex_css_filepath(self.app.conf.resources_dir, relative_from=base_dir),
+            logo=get_firex_logo_filepath(self.app.conf.resources_dir, relative_from=base_dir),
             firex_id=self.app.conf.uid,
             link_for_logo=self.app.conf.link_for_logo,
             header_main_title=self.name_without_orig,
-            worker_log_url=self.worker_log_url,
+            worker_log_url=os.path.relpath(self.worker_log_url, base_dir),
             worker_name=self.request.hostname,
         )
 
@@ -654,6 +679,12 @@ class FireXTask(Task):
         if not self.request.called_directly:
             super(FireXTask, self).send_event(*args, **kwargs)
 
+    def duration(self):
+        return get_time_from_task_start(self.request.id, self.backend)
+
+    def start_time(self):
+        return get_task_start_time(self.request.id, self.backend)
+
 
 def undecorate_func(func):
     undecorated_func = func
@@ -770,3 +801,44 @@ def banner(text, ch='=', length=78, content=''):
     return '\n' + ch * length + '\n' + spaced_text.center(length, ch) + '\n' + content + ch * length + '\n'
 
 
+def get_starttime_dbkey(task_id):
+    return task_id + '_starttime'
+
+
+def get_task_start_time(task_id, backend):
+    starttime_dbkey = get_starttime_dbkey(task_id)
+    try:
+        return float(backend.get(starttime_dbkey))
+    except:
+        return None
+
+
+def get_time_from_task_start(task_id, backend):
+    start_time = get_task_start_time(task_id, backend)
+    if start_time:
+        runtime = time.time() - start_time
+        return runtime
+    return None
+
+
+@task_prerun.connect
+def statsd_task_prerun(sender, task, task_id, args, kwargs, **donotcare):
+    starttime_dbkey = get_starttime_dbkey(task_id)
+    if not sender.backend.get(starttime_dbkey):
+        sender.backend.set(starttime_dbkey, time.time())
+
+
+def send_task_completed_event(task, task_id, backend):
+    actual_runtime = get_time_from_task_start(task_id, backend)
+    if actual_runtime is not None:
+        task.send_event('task-completed', actual_runtime=convert_to_serializable(actual_runtime))
+
+
+@task_postrun.connect
+def statsd_task_postrun(sender, task, task_id, args, kwargs, **donotcare):
+    send_task_completed_event(task, task_id, sender.backend)
+
+
+@task_revoked.connect
+def statsd_task_revoked(sender, request, terminated, signum, expired, **kwargs):
+    send_task_completed_event(sender, request.id, sender.backend)
\ No newline at end of file
